﻿using MySql.Data.MySqlClient;
using Org.BouncyCastle.Asn1.X509.Qualified;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace MontoyaTech.MySqlPlus
{
    /// <summary>
    /// The outline of a MySqlColumn attribute that specify's information about a field
    /// and how it's setup in a MySql Row.
    /// </summary>
    [AttributeUsage(AttributeTargets.Field, AllowMultiple = false, Inherited = true)]
    public class MySqlColumn : Attribute
    {
        /// <summary>
        /// Whether or not this column is the row id.
        /// </summary>
        public bool Id = false;

        /// <summary>
        /// The name of the column, if null, the field name will be used instead.
        /// </summary>
        public string Name = null;

        /// <summary>
        /// An optional MySqlColumnConverter that can convert this column to another
        /// data type when needed.
        /// </summary>
        public Type Converter = null;

        /// <summary>
        /// Creates a new default MySqlColumn.
        /// </summary>
        public MySqlColumn() { }

        /// <summary>
        /// Creates a new MySqlColumn with a name and column converter if needed.
        /// </summary>
        /// <param name="name"></param>
        /// <param name="converter"></param>
        /// <exception cref="NotSupportedException"></exception>
        public MySqlColumn(string name, Type converter = null)
        {
            this.Name = name;

            //Make sure the converter is valid if one was passed.
            if (converter != null && !converter.IsAssignableTo(typeof(MySqlColumnConverter)))
                throw new NotSupportedException($"Converter must inherit {nameof(MySqlColumnConverter)}");

            this.Converter = converter;
        }

        /// <summary>
        /// Reads this column from a given data reader and stores the value into the provided field on a row.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="row"></param>
        /// <param name="field"></param>
        /// <param name="reader"></param>
        /// <exception cref="NotSupportedException"></exception>
        public void ReadValue<T>(T row, FieldInfo field, MySqlDataReader reader)
        {
            //See if we can find the column index.
            int columnIndex = -1;
            int fields = reader.FieldCount;
            for (int i = 0; i < fields; i++)
            {
                string name = reader.GetName(i);

                if (name == this.Name || name == field.Name)
                {
                    columnIndex = i;
                    break;
                }
            }

            //If we didn't find the column, exit.
            if (columnIndex == -1)
                return;

            //See if we have a converter, use it to read the value.
            if (this.Converter != null)
            {
                field.SetValue(row, this.Converter.CreateInstance<MySqlColumnConverter>().ConvertFrom(reader.GetValue(columnIndex)));
            }
            else
            {
                //Get our field type code
                var typeCode = Type.GetTypeCode(field.FieldType);

                //Read the value based on the type code.
                switch (typeCode)
                {
                    case TypeCode.Boolean:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(bool) : reader.GetValue(columnIndex).ConvertToType<bool>());
                        break;
                    case TypeCode.Byte:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(byte) : reader.GetValue(columnIndex).ConvertToType<byte>());
                        break;
                    case TypeCode.Char:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(char) : reader.GetValue(columnIndex).ConvertToType<char>());
                        break;
                    case TypeCode.Decimal:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(decimal) : reader.GetValue(columnIndex).ConvertToType<decimal>());
                        break;
                    case TypeCode.DateTime:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(DateTime) : reader.GetValue(columnIndex).ConvertToType<DateTime>());
                        break;
                    case TypeCode.Double:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(double) : reader.GetValue(columnIndex).ConvertToType<double>());
                        break;
                    case TypeCode.Int16:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(short) : reader.GetValue(columnIndex).ConvertToType<short>());
                        break;
                    case TypeCode.Int32:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(int) : reader.GetValue(columnIndex).ConvertToType<int>());
                        break;
                    case TypeCode.Int64:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(long) : reader.GetValue(columnIndex).ConvertToType<long>());
                        break;
                    case TypeCode.SByte:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(sbyte) : reader.GetValue(columnIndex).ConvertToType<sbyte>());
                        break;
                    case TypeCode.Single:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(float) : reader.GetValue(columnIndex).ConvertToType<float>());
                        break;
                    case TypeCode.String:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(string) : reader.GetValue(columnIndex).ConvertToType<string>());
                        break;
                    case TypeCode.UInt16:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(ushort) : reader.GetValue(columnIndex).ConvertToType<ushort>());
                        break;
                    case TypeCode.UInt32:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(uint) : reader.GetValue(columnIndex).ConvertToType<uint>());
                        break;
                    case TypeCode.UInt64:
                        field.SetValue(row, reader.IsDBNull(columnIndex) ? default(ulong) : reader.GetValue(columnIndex).ConvertToType<ulong>());
                        break;
                    default:
                        throw new NotSupportedException("Unsupported TypeCode");
                }
            }
        }
    }
}
